/**********************************************************************

Drive the left wheel with sinusoidal commands.
Tell time by polling a clock instead of using interrupts.
Use a Kalman-filter-based velocity estimator to restore signs to the
 encoder signal.
Print out data concurrently for capture by a terminal program (putty).

Collecting data is organized into "trials".
During a trial, the effect of the sinusoidal drive is gradually
increased with a ramp at the beginning, and gradually decreased with a
ramp at the end.

[Ignore] means that you can safely ignore this item in trying to figure
out how this program works.

***********************************************************************/

#include <math.h>

/**********************************************************************/
// Global variables

// motor enable
bool go = false;

// The encoder interrupt service routines (ISRs) update these variables
volatile unsigned long int raw_right_encoder_count = 0;
volatile unsigned long int raw_left_encoder_count = 0;

// Motor commands that have been sent to the motors
int left_command;
int right_command;

// Velocity estimation
// used in velocity estimation
unsigned long int last_raw_left_encoder_count = 0;
// estimated wheel velocity
float left_wheel_velocity_encoder_estimated = 0;
float left_wheel_position_encoder_estimated = 0;
float a_vel_est = 0.9325;
float b_vel_est = 0.0045;
float k_vel_est = 0.1;

int my_debug = 0; // useful for debugging

/**********************************************************************/

#include "ignore.h" // [Ignore]

/**********************************************************************/

unsigned long int read_left_encoder()
{

  unsigned long int rlec1 = raw_left_encoder_count; // try reading.
  unsigned long int rlec2 = raw_left_encoder_count; // try reading again.
  unsigned long int rlec3 = raw_left_encoder_count; // one more time.
  
  my_debug = 0;

  // try 1st pair
  if ( rlec1 == rlec2 ) // readings agree, go with them.
    {
      my_debug = 1;
      return rlec1;
    }

  // try 2nd pair
  if ( rlec2 == rlec3 ) // readings agree, go with them.
    {
      my_debug = 2;
      return rlec2;
    }

  // find median
  if ( rlec1 < rlec2 )
    {
      if ( rlec2 < rlec3 )
        { // rlec1 < rlec2 < rlec3
          my_debug = 3;
	  return rlec2;
        }
      else
        {
          // rlec1 < rlec2 && rlec3 < rlec2
          if ( rlec1 < rlec3 )
	    { // rlec1 < rlec3 < rlec2
	      my_debug = 4;
	      return rlec3;
	    }
          else
	    { // rlec3 < rlec1 < rlec2
	      my_debug = 5;
	      return rlec1;
	    }
        }
    }
  else // rlec2 < rlec1
    {
      if ( rlec1 < rlec3 )
        { // rlec2 < rlec1 < rlec3
          my_debug = 6;
	  return rlec1;
        }
      else
        { // rlec2 < rlec1 && rlec3 < rlec1
          if ( rlec2 < rlec3 )
	    { // rlec2 < rlec3 < rlec1
	      my_debug = 7;
	      return rlec3;
	    }
          else
	    { // rlec3 < rlec2 < rlec1
	      my_debug = 8;
	      return rlec2;
	    }
        }
    }

  // should never get here
  my_debug = 9;
  return rlec3;
}     

/**********************************************************************/
// note that int arguments have to be less than +/-32000

void run_trial( float frequency,
                unsigned long run_time_limit,
		int collect_data )
{

#include "run_trial1.h" // [Ignore]

  // print out the frequency
  if ( collect_data )
    {
      Serial.print( "Frequency " );
      Serial.println( frequency );
    }

  // Initialize velocity estimator
  last_raw_left_encoder_count = raw_left_encoder_count;
  left_wheel_velocity_encoder_estimated = 0.0;
  left_wheel_position_encoder_estimated = 0.0;

  // servo loop
  for( ; ; )
    {

#include "run_trial2.h" // [Ignore]

/*****************************************
State estimation part */

      float left_prediction
      	    = a_vel_est*left_wheel_velocity_encoder_estimated
      	    + b_vel_est*left_command;

      unsigned long int the_raw_left_encoder_count = read_left_encoder();

      int left_increment  // this is always positive or zero
      	  = the_raw_left_encoder_count - last_raw_left_encoder_count;
      last_raw_left_encoder_count = the_raw_left_encoder_count;

      int left_measured = 0;
      if ( left_prediction > 0 )
        left_measured = left_increment;
      else if ( left_prediction < 0 )
        left_measured = -left_increment;
      else if ( left_command > 0 )
        left_measured = left_increment;
      else if ( left_command < 0 )
        left_measured = -left_increment;
      else if ( left_wheel_velocity_encoder_estimated > 0 )
        left_measured = left_increment;
      else if ( left_wheel_velocity_encoder_estimated < 0 )
        left_measured = -left_increment;
      else
        left_measured = left_increment; // ran out of criteria to guide this
      left_wheel_velocity_encoder_estimated
        = left_prediction + k_vel_est*(left_measured - left_prediction);
      left_wheel_position_encoder_estimated
        += left_wheel_velocity_encoder_estimated;

      /*
      There are several alternative ways to compute the actual velocity,
      once we have assigned signs (positive or negative) to the increments.
      left_wheel_velocity_encoder_estimated is a real value based on the
      Kalman filter, and left_wheel_position_encoder_estimated is the sum
      of the left_wheel_velocity_encoder_estimated values.
      The increments themselves are too noisy at low velocities
      (since they have to be integers) to use directly as velocity estimates.
      We can, however, sum up the signed increments to estimate the total
      wheel angle in encoder counts. This would probably work fine.
      We are currently assuming the servo ticks are perfectly regular.
      We should be dividing the increments by the actual time since the
      last reading, to compensate for variation in when the encoders are
      actually read.
      */

/*****************************************
Control part */

      float left_commandf = 0;
      
      left_commandf = MAX_COMMAND*sinf( frequency*(2*PI*0.001)*run_time + PI/2);
	  // (2*PI*0.001) makes frequency in Hz

      // ramp command up at beginning
      if ( ( run_time < START_COLLECT )
           && ( ramp < (1.0 - RAMP_INCREMENT) ) )
        ramp += RAMP_INCREMENT;

      // ramp command down at end
      if ( run_time > run_time_limit )
        ramp *= RAMP_TURNOFF_FACTOR; // gradually turn motor off

      int left_cmmnd = round( ramp*left_commandf );

      motor_left_command( left_cmmnd );

/*****************************************
Data printing part */

      // print out debugging info before data collection
      if ( ( debug_print_time > DEBUG_PRINT_INTERVAL )
           && ( ( run_time < START_COLLECT )
	        || ( run_time > run_time_limit ) ) )
        {
          ProcessCommand();
          debug_print_time -= DEBUG_PRINT_INTERVAL;
	  // you could print stuff out here.
        }

      // print out data
      if ( ( run_time >= START_COLLECT ) && ( run_time <= run_time_limit )
           && collect_data )
        {
          if ( !started )
            {
              servo_late = 0;
              max_servo_late = 0;
              started = 1;
              Serial.println( "Data" );
            }
          Serial.print( current_micros );
          Serial.print( " " );
          Serial.print( the_raw_left_encoder_count );
          Serial.print( " " );
          Serial.print( left_command );
          Serial.print( " " );
          Serial.print( round(1000*left_wheel_velocity_encoder_estimated) );
          Serial.print( " " );
          Serial.print( round(10*left_wheel_position_encoder_estimated) );
          Serial.print( " " );
          Serial.print( round(1000*left_prediction) );
          Serial.print( " " );
          Serial.print( left_measured );
          Serial.print( " " );
          Serial.println( my_debug );
        }
    }
}

/**********************************************************************/
// this subroutine is repeatedly called

void loop()
{

  // if motor is not enabled check for a command
  if ( !go )
    {
      motor_stop();
      delay(500);
      ProcessCommand();
      return;
    }

  // run trials if motor enabled
  run_trial( 0.03, // frequency
	     100000, // Run time limit in milliseconds
	     1 ); // Should data be collected?
  delay(1000);

  run_trial( 0.05, // frequency
	     70000, // Run time limit in milliseconds
	     1 ); // Should data be collected?
  delay(1000);

  run_trial( 0.1, // frequency
	     60000, // Run time limit in milliseconds
	     1 ); // Should data be collected?
  delay(1000);

  run_trial( 0.2, // frequency
	     50000, // Run time limit in milliseconds
	     1 ); // Should data be collected?
  delay(1000);

  run_trial( 0.3, // frequency
	     40000, // Run time limit in milliseconds
	     1 ); // Should data be collected?
  delay(1000);

  run_trial( 0.5, // frequency
	     30000, // Run time limit in milliseconds
	     1 ); // Should data be collected?
  delay(1000);

  run_trial( 0.7, // frequency
	     20000, // Run time limit in milliseconds
	     1 ); // Should data be collected?
  delay(1000);

  run_trial( 1.0, // frequency
	     10000, // Run time limit in milliseconds
	     1 ); // Should data be collected?
  delay(1000);

  run_trial( 2.0, // frequency
	     10000, // Run time limit in milliseconds
	     1 ); // Should data be collected?
  delay(1000);

  run_trial( 3.0, // frequency
	     10000, // Run time limit in milliseconds
	     1 ); // Should data be collected?
  delay(1000);

  run_trial( 4.0, // frequency
	     10000, // Run time limit in milliseconds
	     1 ); // Should data be collected?
  delay(1000);

  run_trial( 5.0, // frequency
	     10000, // Run time limit in milliseconds
	     1 ); // Should data be collected?
  delay(1000);

  run_trial( 6.0, // frequency
	     10000, // Run time limit in milliseconds
	     1 ); // Should data be collected?
  delay(1000);

  run_trial( 7.0, // frequency
	     10000, // Run time limit in milliseconds
	     1 ); // Should data be collected?
  delay(1000);

  run_trial( 8.0, // frequency
	     10000, // Run time limit in milliseconds
	     1 ); // Should data be collected?
  delay(1000);

  run_trial( 9.0, // frequency
	     10000, // Run time limit in milliseconds
	     1 ); // Should data be collected?
  delay(1000);

  run_trial( 10.0, // frequency
	     10000, // Run time limit in milliseconds
	     1 ); // Should data be collected?
  delay(1000);

  run_trial( 11.0, // frequency
	     10000, // Run time limit in milliseconds
	     1 ); // Should data be collected?
  delay(1000);

  run_trial( 12.0, // frequency
	     10000, // Run time limit in milliseconds
	     1 ); // Should data be collected?
  delay(1000);

  run_trial( 13.0, // frequency
	     10000, // Run time limit in milliseconds
	     1 ); // Should data be collected?
  delay(1000);

  run_trial( 14.0, // frequency
	     10000, // Run time limit in milliseconds
	     1 ); // Should data be collected?
  delay(1000);

  run_trial( 15.0, // frequency
	     10000, // Run time limit in milliseconds
	     1 ); // Should data be collected?
  delay(1000);

  run_trial( 16.0, // frequency
	     10000, // Run time limit in milliseconds
	     1 ); // Should data be collected?
  delay(1000);

  run_trial( 17.0, // frequency
	     10000, // Run time limit in milliseconds
	     1 ); // Should data be collected?
  delay(1000);

  run_trial( 18.0, // frequency
	     10000, // Run time limit in milliseconds
	     1 ); // Should data be collected?
  delay(1000);

  run_trial( 19.0, // frequency
	     10000, // Run time limit in milliseconds
	     1 ); // Should data be collected?
  delay(1000);

  run_trial( 20.0, // frequency
	     10000, // Run time limit in milliseconds
	     1 ); // Should data be collected?
  delay(1000);


  // disable the motor after a trial
  Serial.println( "Stop!" );
  go = false;
}
  
/**********************************************************************/
